// ===== Wolfgang Erben ========== Modellierung =====
package pruefung2;

import static java.lang.Math.*;
import studienarbeit.IPunkt;

/**
 * beschreibt eine Stelle in der Ebene.
 * Objekte der Klasse sind unveraenderbar und verhalten sich kanonisch. 
 *
 * @author Wolfgang Erben
 * @version 1.3 03.06.2010
 */
public final class Punkt implements java.io.Serializable, IPunkt { 
    private static final long serialVersionUID = 1L;
    private final double mX, mY;
    public static final Punkt URSPRUNG = new Punkt(0., 0.);
    private static final double HALB_RUM = 180.;
    private static final double FAKTOR = PI / HALB_RUM;
    private static final int HASH_BASIS = 17;
    private static final int HASH_FAKTOR = 31;
    
    /**
     * erzeugt einen Punkt mit den angegebenen Koordinaten.
     */
    public Punkt(double x, double y) {
        mX = x;
        mY = y;
    }
    /**
     * erzeugt den Ursprung.
     */
    public Punkt() {
        this(0., 0.);
    }
    /**
     * erzeugt einen Punkt mit den gleichen Koordinaten
     * wie der angegebene.
     */
    public Punkt(IPunkt p) {
        mX = p.x();
        mY = p.y();
    }
    /**
     * liefert die x-Koordinate dieses Punktes.
     */
    @Override
    public double x() {
        return mX;
    }
    
    /**
     * liefert die y-Koordinate dieses Punktes.
     */
    @Override
    public double y() {
        return mY;
    }
    
    /**
     * liefert den Punkt mit den baryzentrischen Koordinaten
     * 1-t und t bezueglich diesem Punkt und dem angegebenen.
     * Fuer t=0 ergibt sich also dieser Punkt, fuer t=1 der angegebene.
     */
    public Punkt inRichtung(IPunkt anderer, double t) {
        double x = anderer.x() * t + mX * (1. - t);
        double y = anderer.y() * t + mY * (1. - t);
        return new Punkt(x, y);
    }
    /** 
     * liefert die Mitte zwischen diesem Punkt und dem angegebenen.
     */
    public Punkt mitteZu(IPunkt anderer) {
        final double halb = 0.5;
        return inRichtung(anderer, halb);
    }
    /**
     * liefert den Punkt, der durch Drehung des angegebenen Punktes
     * um den angegebenen Winkel entsteht. Der Winkel ist im Gradmass
     * anzugeben. Drehpunkt ist der Ursprung.
     */
    public Punkt gedrehtUm(double winkel) {
        double bogen = FAKTOR * winkel;
        double x = mX * cos(bogen) - mY * sin(bogen);
        double y = mX * sin(bogen) + mY * cos(bogen);
        return new Punkt(x, y);
    }
    
    /**
     * liefert den Abstand des Punktes zum angegebenen Punkt.
     */
    public double abstandsQuadratZu(IPunkt p) {
        return abstandsQuadratZu(p.x(), p.y());
    }
    /**
     * liefert den Abstand des Punktes zum angegebenen Punkt.
     */
    @Override
    public double abstandZu(IPunkt p) {
        return abstandZu(p.x(), p.y());
    }
    /**
     * liefert den Abstand des Punktes zum angegebenen Punkt.
     */
    public double abstandsQuadratZu(double x, double y) {
        double dx = x - x(); 
        double dy = y - y();
        return dx * dx + dy * dy;
    }
    /**
     * liefert den Abstand des Punktes zum angegebenen Punkt.
     */
    public double abstandZu(double x, double y) {
        double d2 = abstandsQuadratZu(x, y);
        return Math.sqrt(d2);
    }
    
    /**
     * liefert eine textuelle Beschreibung dieses Punktes.
     */
    @Override
    public String toString() {
        return "(" + mX + "," + mY + ")";
    }
    
    /**
     * liefert einen mit <code>equals</code> kompatiblen Hash-Code.
     */
    @Override
    public int hashCode() {
        int result = HASH_BASIS;
        result = HASH_FAKTOR * result + new Double(mX).hashCode();
        result = HASH_FAKTOR * result + new Double(mY).hashCode();
        return result;
    }
    
    /**
     * prueft, ob dieser und der angegebene Punkt die gleichen
     * Koordinaten besitzen.
     */
    @Override
    public boolean equals(Object anderes) {
        if (this == anderes) return true;
        if (anderes == null) return false;
        if (!(anderes instanceof Punkt)) return false;
        Punkt p = (Punkt) anderes;
        if (p.mX != mX) return false;
        if (p.mY != mY) return false;
        return true;
    }

}
